@@ -325,6 +325,26 @@ ADD_ALPHA_FUNC(r8g8b8_to_b8g8r8a8, 0, 1, 2, 2, 1, 0, 3)
 ADD_ALPHA_FUNC(r8g8b8_to_a8r8g8b8, 0, 1, 2, 1, 2, 3, 0)
 ADD_ALPHA_FUNC(r8g8b8_to_a8b8g8r8, 0, 1, 2, 3, 2, 1, 0)
 
+#define SWAP_FUNC(name, R, G, B, A) \
+static void \
+name (guchar *dest, \
+      const guchar *src, \
+      gsize n) \
+{ \
+  for (; n > 0; n--) \
+    { \
+      dest[0] = src[R]; \
+      dest[1] = src[G]; \
+      dest[2] = src[B]; \
+      dest[3] = src[A]; \
+      dest += 4; \
+      src += 4; \
+    } \
+}
+
+SWAP_FUNC(r8g8b8a8_to_b8g8r8a8, 2, 1, 0, 3)
+SWAP_FUNC(b8g8r8a8_to_r8g8b8a8, 2, 1, 0, 3)
+
 #define MIPMAP_FUNC(SumType, DataType, n_units) \
 static void \
 gdk_mipmap_ ## DataType ## _ ## n_units ## _nearest (guchar       *dest, \
@@ -2032,6 +2052,12 @@ get_fast_conversion_func (GdkMemoryFormat dest_format,
     return r8g8b8a8_to_a8r8g8b8_premultiplied;
   else if (src_format == GDK_MEMORY_B8G8R8A8 && dest_format == GDK_MEMORY_A8R8G8B8_PREMULTIPLIED)
     return r8g8b8a8_to_a8b8g8r8_premultiplied;
+  else if ((src_format == GDK_MEMORY_B8G8R8A8 && dest_format == GDK_MEMORY_R8G8B8A8) ||
+           (src_format == GDK_MEMORY_B8G8R8A8_PREMULTIPLIED && dest_format == GDK_MEMORY_R8G8B8A8_PREMULTIPLIED))
+    return b8g8r8a8_to_r8g8b8a8;
+  else if ((src_format == GDK_MEMORY_R8G8B8A8 && dest_format == GDK_MEMORY_B8G8R8A8) ||
+           (src_format == GDK_MEMORY_R8G8B8A8_PREMULTIPLIED && dest_format == GDK_MEMORY_B8G8R8A8_PREMULTIPLIED))
+    return r8g8b8a8_to_b8g8r8a8;
   else if (src_format == GDK_MEMORY_R8G8B8 && dest_format == GDK_MEMORY_R8G8B8A8_PREMULTIPLIED)
     return r8g8b8_to_r8g8b8a8;
   else if (src_format == GDK_MEMORY_B8G8R8 && dest_format == GDK_MEMORY_R8G8B8A8_PREMULTIPLIED)
@@ -2074,6 +2100,7 @@ struct _MemoryConvert
   GdkColorState       *src_cs;
   gsize                width;
   gsize                height;
+  gsize                chunk_size;
 
   /* atomic */ int     rows_done;
 };
@@ -2088,20 +2115,50 @@ gdk_memory_convert_generic (gpointer data)
   GdkFloatColorConvert convert_func = NULL;
   GdkFloatColorConvert convert_func2 = NULL;
   gboolean needs_premultiply, needs_unpremultiply;
-  gsize y, n;
+  gsize y0, y;
   gint64 before = GDK_PROFILER_CURRENT_TIME;
   gsize rows;
 
-  convert_func = gdk_color_state_get_convert_to (mc->src_cs, mc->dest_cs);
-
-  if (!convert_func)
-    convert_func2 = gdk_color_state_get_convert_from (mc->dest_cs, mc->src_cs);
-
-  if (!convert_func && !convert_func2)
+  if (gdk_color_state_equal (mc->src_cs, mc->dest_cs))
     {
-      GdkColorState *connection = GDK_COLOR_STATE_REC2100_LINEAR;
-      convert_func = gdk_color_state_get_convert_to (mc->src_cs, connection);
-      convert_func2 = gdk_color_state_get_convert_from (mc->dest_cs, connection);
+      FastConversionFunc func;
+
+      func = get_fast_conversion_func (mc->dest_format, mc->src_format);
+
+      if (func != NULL)
+        {
+          for (y0 = g_atomic_int_add (&mc->rows_done, mc->chunk_size), rows = 0;
+               y0 < mc->height;
+               y0 = g_atomic_int_add (&mc->rows_done, mc->chunk_size))
+            {
+              for (y = y0; y < MIN (y0 + mc->chunk_size, mc->height); y++, rows++)
+                {
+                  const guchar *src_data = mc->src_data + y * mc->src_stride;
+                  guchar *dest_data = mc->dest_data + y * mc->dest_stride;
+
+                  func (dest_data, src_data, mc->width);
+                }
+            }
+
+          ADD_MARK (before,
+                    "Memory convert (thread)", "size %lux%lu, %lu rows",
+                    mc->width, mc->height, rows);
+          return;
+        }
+    }
+  else
+    {
+      convert_func = gdk_color_state_get_convert_to (mc->src_cs, mc->dest_cs);
+
+      if (!convert_func)
+        convert_func2 = gdk_color_state_get_convert_from (mc->dest_cs, mc->src_cs);
+
+      if (!convert_func && !convert_func2)
+        {
+          GdkColorState *connection = GDK_COLOR_STATE_REC2100_LINEAR;
+          convert_func = gdk_color_state_get_convert_to (mc->src_cs, connection);
+          convert_func2 = gdk_color_state_get_convert_from (mc->dest_cs, connection);
+        }
     }
 
   if (convert_func)
@@ -2116,30 +2173,32 @@ gdk_memory_convert_generic (gpointer data)
     }
 
   tmp = g_malloc (sizeof (*tmp) * mc->width);
-  n = 1;
 
-  for (y = g_atomic_int_add (&mc->rows_done, n), rows = 0;
-       y < mc->height;
-       y = g_atomic_int_add (&mc->rows_done, n), rows++)
+  for (y0 = g_atomic_int_add (&mc->rows_done, mc->chunk_size), rows = 0;
+       y0 < mc->height;
+       y0 = g_atomic_int_add (&mc->rows_done, mc->chunk_size))
     {
-      const guchar *src_data = mc->src_data + y * mc->src_stride;
-      guchar *dest_data = mc->dest_data + y * mc->dest_stride;
+      for (y = y0; y < MIN (y0 + mc->chunk_size, mc->height); y++, rows++)
+        {
+          const guchar *src_data = mc->src_data + y * mc->src_stride;
+          guchar *dest_data = mc->dest_data + y * mc->dest_stride;
 
-      src_desc->to_float (tmp, src_data, mc->width);
+          src_desc->to_float (tmp, src_data, mc->width);
 
-      if (needs_unpremultiply)
-        unpremultiply (tmp, mc->width);
+          if (needs_unpremultiply)
+            unpremultiply (tmp, mc->width);
 
-      if (convert_func)
-        convert_func (mc->src_cs, tmp, mc->width);
+          if (convert_func)
+            convert_func (mc->src_cs, tmp, mc->width);
 
-      if (convert_func2)
-        convert_func2 (mc->dest_cs, tmp, mc->width);
+          if (convert_func2)
+            convert_func2 (mc->dest_cs, tmp, mc->width);
 
-      if (needs_premultiply)
-        premultiply (tmp, mc->width);
+          if (needs_premultiply)
+            premultiply (tmp, mc->width);
 
-      dest_desc->from_float (dest_data, tmp, mc->width);
+          dest_desc->from_float (dest_data, tmp, mc->width);
+        }
     }
 
   g_free (tmp);
@@ -2172,8 +2231,11 @@ gdk_memory_convert (guchar              *dest_data,
     .src_cs = src_cs,
     .width = width,
     .height = height,
+    .chunk_size = MAX (1, 512 / width),
   };
 
+  guint n_tasks;
+
   g_assert (dest_format < GDK_MEMORY_N_FORMATS);
   g_assert (src_format < GDK_MEMORY_N_FORMATS);
   /* We don't allow overlap here. If you want to do in-place color state conversions,
@@ -2205,27 +2267,9 @@ gdk_memory_convert (guchar              *dest_data,
       return;
     }
 
-  if (gdk_color_state_equal (dest_cs, src_cs))
-    {
-      FastConversionFunc func;
+  n_tasks = (mc.height + mc.chunk_size - 1) / mc.chunk_size;
 
-      func = get_fast_conversion_func (dest_format, src_format);
-
-      if (func != NULL)
-        {
-          gsize y;
-
-          for (y = 0; y < height; y++)
-            {
-              func (dest_data, src_data, width);
-              src_data += src_stride;
-              dest_data += dest_stride;
-            }
-          return;
-        }
-    }
-
-  gdk_parallel_task_run (gdk_memory_convert_generic, &mc);
+  gdk_parallel_task_run (gdk_memory_convert_generic, &mc, n_tasks);
 }
 
 typedef struct _MemoryConvertColorState MemoryConvertColorState;
@@ -2239,6 +2283,7 @@ struct _MemoryConvertColorState
   GdkColorState *dest_cs;
   gsize width;
   gsize height;
+  gsize chunk_size;
 
   /* atomic */ int rows_done;
 };
@@ -2350,15 +2395,16 @@ static void
 gdk_memory_convert_color_state_srgb_to_srgb_linear (gpointer data)
 {
   MemoryConvertColorState *mc = data;
-  int y;
+  gsize y0, y;
   guint64 before = GDK_PROFILER_CURRENT_TIME;
   gsize rows;
 
-  for (y = g_atomic_int_add (&mc->rows_done, 1), rows = 0;
-       y < mc->height;
-       y = g_atomic_int_add (&mc->rows_done, 1), rows++)
+  for (y0 = g_atomic_int_add (&mc->rows_done, mc->chunk_size), rows = 0;
+       y0 < mc->height;
+       y0 = g_atomic_int_add (&mc->rows_done, mc->chunk_size))
     {
-      convert_srgb_to_srgb_linear (mc->data + y * mc->stride, mc->width);
+      for (y = y0; y < MIN (y0 + mc->chunk_size, mc->height); y++, rows++)
+        convert_srgb_to_srgb_linear (mc->data + y * mc->stride, mc->width);
     }
 
   ADD_MARK (before,
@@ -2370,15 +2416,16 @@ static void
 gdk_memory_convert_color_state_srgb_linear_to_srgb (gpointer data)
 {
   MemoryConvertColorState *mc = data;
-  int y;
+  gsize y0, y;
   guint64 before = GDK_PROFILER_CURRENT_TIME;
   gsize rows;
 
-  for (y = g_atomic_int_add (&mc->rows_done, 1), rows = 0;
-       y < mc->height;
-       y = g_atomic_int_add (&mc->rows_done, 1), rows++)
+  for (y0 = g_atomic_int_add (&mc->rows_done, mc->chunk_size), rows = 0;
+       y0 < mc->height;
+       y0 = g_atomic_int_add (&mc->rows_done, mc->chunk_size))
     {
-      convert_srgb_linear_to_srgb (mc->data + y * mc->stride, mc->width);
+      for (y = y0; y < MIN (y0 + mc->chunk_size, mc->height); y++, rows++)
+        convert_srgb_linear_to_srgb (mc->data + y * mc->stride, mc->width);
     }
 
   ADD_MARK (before,
@@ -2394,7 +2441,7 @@ gdk_memory_convert_color_state_generic (gpointer user_data)
   GdkFloatColorConvert convert_func = NULL;
   GdkFloatColorConvert convert_func2 = NULL;
   float (*tmp)[4];
-  int y;
+  gsize y0, y;
   guint64 before = GDK_PROFILER_CURRENT_TIME;
   gsize rows;
 
@@ -2414,27 +2461,30 @@ gdk_memory_convert_color_state_generic (gpointer user_data)
 
   tmp = g_malloc (sizeof (*tmp) * mc->width);
 
-  for (y = g_atomic_int_add (&mc->rows_done, 1), rows = 0;
-       y < mc->height;
-       y = g_atomic_int_add (&mc->rows_done, 1), rows++)
+  for (y0 = g_atomic_int_add (&mc->rows_done, mc->chunk_size), rows = 0;
+       y0 < mc->height;
+       y0 = g_atomic_int_add (&mc->rows_done, mc->chunk_size))
     {
-      guchar *data = mc->data + y * mc->stride;
+      for (y = y0; y < MIN (y0 + mc->chunk_size, mc->height); y++, rows++)
+        {
+          guchar *data = mc->data + y * mc->stride;
 
-      desc->to_float (tmp, data, mc->width);
+          desc->to_float (tmp, data, mc->width);
 
-      if (desc->alpha == GDK_MEMORY_ALPHA_PREMULTIPLIED)
-        unpremultiply (tmp, mc->width);
+          if (desc->alpha == GDK_MEMORY_ALPHA_PREMULTIPLIED)
+            unpremultiply (tmp, mc->width);
 
-      if (convert_func)
-        convert_func (mc->src_cs, tmp, mc->width);
+          if (convert_func)
+            convert_func (mc->src_cs, tmp, mc->width);
 
-      if (convert_func2)
-        convert_func2 (mc->dest_cs, tmp, mc->width);
+          if (convert_func2)
+            convert_func2 (mc->dest_cs, tmp, mc->width);
 
-      if (desc->alpha == GDK_MEMORY_ALPHA_PREMULTIPLIED)
-        premultiply (tmp, mc->width);
+          if (desc->alpha == GDK_MEMORY_ALPHA_PREMULTIPLIED)
+            premultiply (tmp, mc->width);
 
-      desc->from_float (data, tmp, mc->width);
+          desc->from_float (data, tmp, mc->width);
+        }
     }
 
   g_free (tmp);
@@ -2448,8 +2498,8 @@ void
 gdk_memory_convert_color_state (guchar          *data,
                                 gsize            stride,
                                 GdkMemoryFormat  format,
-                                GdkColorState   *src_cs,
-                                GdkColorState   *dest_cs,
+                                GdkColorState   *src_color_state,
+                                GdkColorState   *dest_color_state,
                                 gsize            width,
                                 gsize            height)
 {
@@ -2457,30 +2507,34 @@ gdk_memory_convert_color_state (guchar          *data,
     .data = data,
     .stride = stride,
     .format = format,
-    .src_cs = src_cs,
-    .dest_cs = dest_cs,
+    .src_cs = src_color_state,
+    .dest_cs = dest_color_state,
     .width = width,
     .height = height,
+    .chunk_size = MAX (1, 512 / width),
   };
+  guint n_tasks;
 
-  if (gdk_color_state_equal (src_cs, dest_cs))
+  if (gdk_color_state_equal (src_color_state, dest_color_state))
     return;
 
+  n_tasks = (mc.height + mc.chunk_size - 1) / mc.chunk_size;
+
   if (format == GDK_MEMORY_B8G8R8A8_PREMULTIPLIED &&
-      src_cs == GDK_COLOR_STATE_SRGB &&
-      dest_cs == GDK_COLOR_STATE_SRGB_LINEAR)
+      src_color_state == GDK_COLOR_STATE_SRGB &&
+      dest_color_state == GDK_COLOR_STATE_SRGB_LINEAR)
     {
-      gdk_parallel_task_run (gdk_memory_convert_color_state_srgb_to_srgb_linear, &mc);
+      gdk_parallel_task_run (gdk_memory_convert_color_state_srgb_to_srgb_linear, &mc, n_tasks);
     }
   else if (format == GDK_MEMORY_B8G8R8A8_PREMULTIPLIED &&
-           src_cs == GDK_COLOR_STATE_SRGB_LINEAR &&
-           dest_cs == GDK_COLOR_STATE_SRGB)
+           src_color_state == GDK_COLOR_STATE_SRGB_LINEAR &&
+           dest_color_state == GDK_COLOR_STATE_SRGB)
     {
-      gdk_parallel_task_run (gdk_memory_convert_color_state_srgb_linear_to_srgb, &mc);
+      gdk_parallel_task_run (gdk_memory_convert_color_state_srgb_linear_to_srgb, &mc, n_tasks);
     }
   else
     {
-      gdk_parallel_task_run (gdk_memory_convert_color_state_generic, &mc);
+      gdk_parallel_task_run (gdk_memory_convert_color_state_generic, &mc, n_tasks);
     }
 }
 
@@ -2636,19 +2690,24 @@ gdk_memory_mipmap (guchar          *dest,
     .linear = linear,
     .rows_done = 0,
   };
+  gsize chunk_size;
+  guint n_tasks;
 
   g_assert (lod_level > 0);
 
+  chunk_size = MAX (1, 512 / src_width),
+  n_tasks = (src_height + chunk_size - 1) / chunk_size;
+
   if (dest_format == src_format)
     {
       if (linear)
-        gdk_parallel_task_run (gdk_memory_mipmap_same_format_linear, &mipmap);
+        gdk_parallel_task_run (gdk_memory_mipmap_same_format_linear, &mipmap, n_tasks);
       else
-        gdk_parallel_task_run (gdk_memory_mipmap_same_format_nearest, &mipmap);
+        gdk_parallel_task_run (gdk_memory_mipmap_same_format_nearest, &mipmap, n_tasks);
     }
   else
     {
-      gdk_parallel_task_run (gdk_memory_mipmap_generic, &mipmap);
+      gdk_parallel_task_run (gdk_memory_mipmap_generic, &mipmap, n_tasks);
     }
 }
 
