@@ -35,21 +35,21 @@
 #include <math.h>
 
 #ifdef GDK_RENDERING_VULKAN
-static const GdkDebugKey gdk_vulkan_feature_keys[] = {
+const GdkDebugKey gdk_vulkan_feature_keys[] = {
   { "dmabuf", GDK_VULKAN_FEATURE_DMABUF, "Never import Dmabufs" },
   { "ycbcr", GDK_VULKAN_FEATURE_YCBCR, "Do not support Ycbcr textures (also disables dmabufs)" },
   { "semaphore-export", GDK_VULKAN_FEATURE_SEMAPHORE_EXPORT, "Disable sync of exported dmabufs" },
   { "semaphore-import", GDK_VULKAN_FEATURE_SEMAPHORE_IMPORT, "Disable sync of imported dmabufs" },
   { "incremental-present", GDK_VULKAN_FEATURE_INCREMENTAL_PRESENT, "Do not send damage regions" },
   { "swapchain-maintenance", GDK_VULKAN_FEATURE_SWAPCHAIN_MAINTENANCE, "Do not use advanced swapchain features" },
+  { "swapchain-colorspace", GDK_VULKAN_FEATURE_SWAPCHAIN_COLORSPACE, "Force default colorspace (likely crashes Wayland)" },
 };
 #endif
 
 /**
  * GdkVulkanContext:
  *
- * `GdkVulkanContext` is an object representing the platform-specific
- * Vulkan draw context.
+ * Represents the platform-specific Vulkan draw context.
  *
  * `GdkVulkanContext`s are created for a surface using
  * [method@Gdk.Surface.create_vulkan_context], and the context will match
@@ -300,6 +300,10 @@ surface_present_mode_to_string (VkPresentModeKHR present_mode)
       return "VK_PRESENT_MODE_FIFO_KHR";
     case VK_PRESENT_MODE_FIFO_RELAXED_KHR:
       return "VK_PRESENT_MODE_FIFO_RELAXED_KHR";
+#if VK_HEADER_VERSION >= 304
+    case VK_PRESENT_MODE_FIFO_LATEST_READY_EXT:
+      return "VK_PRESENT_MODE_FIFO_LATEST_READY_EXT";
+#endif
     case VK_PRESENT_MODE_SHARED_DEMAND_REFRESH_KHR:
     case VK_PRESENT_MODE_SHARED_CONTINUOUS_REFRESH_KHR:
     case VK_PRESENT_MODE_MAX_ENUM_KHR:
@@ -367,7 +371,6 @@ gdk_vulkan_context_dispose (GObject *gobject)
 {
   GdkVulkanContext *context = GDK_VULKAN_CONTEXT (gobject);
   GdkVulkanContextPrivate *priv = gdk_vulkan_context_get_instance_private (context);
-  GdkDisplay *display;
   VkDevice device;
   guint i;
 
@@ -397,11 +400,6 @@ gdk_vulkan_context_dispose (GObject *gobject)
       priv->surface = VK_NULL_HANDLE;
     }
 
-  /* display will be unset in gdk_draw_context_dispose() */
-  display = gdk_draw_context_get_display (GDK_DRAW_CONTEXT (context));
-  if (display && priv->vulkan_ref)
-    gdk_display_unref_vulkan (display);
-
   G_OBJECT_CLASS (gdk_vulkan_context_parent_class)->dispose (gobject);
 }
 
@@ -417,6 +415,7 @@ gdk_vulkan_context_check_swapchain (GdkVulkanContext  *context,
   VkSwapchainKHR new_swapchain;
   VkResult res;
   VkDevice device;
+  VkExtent2D size;
   guint i;
 
   GDK_DEBUG (VULKAN, "(Re)creating the swapchain for surface of size %dx%d",
@@ -460,10 +459,8 @@ gdk_vulkan_context_check_swapchain (GdkVulkanContext  *context,
 
   GDK_DEBUG (VULKAN, "Using surface present mode %s",
              surface_present_mode_to_string (present_mode));
-  GDK_DEBUG (VULKAN, "Using extent %dx%d",
-             capabilities.currentExtent.width,
-             capabilities.currentExtent.height);
 
+  gdk_draw_context_get_buffer_size (GDK_DRAW_CONTEXT (context), &size.width, &size.height);
 
   /*
    * Per https://www.khronos.org/registry/vulkan/specs/1.0-wsi_extensions/xhtml/vkspec.html#VkSurfaceCapabilitiesKHR
@@ -472,14 +469,34 @@ gdk_vulkan_context_check_swapchain (GdkVulkanContext  *context,
    */
   if (capabilities.currentExtent.width == -1 || capabilities.currentExtent.height == -1)
     {
-      double scale = gdk_surface_get_scale (surface);
+      GDK_DEBUG (VULKAN, "GTK selecting swapchain size as %ux%u",
+                 (guint) size.width, (guint) size.height);
+    }
+  else if (capabilities.currentExtent.width == size.width &&
+           capabilities.currentExtent.height == size.height)
+    {
+      GDK_DEBUG (VULKAN, "Vulkan and GTK agree on size as %ux%u",
+                 (guint) size.width, (guint) size.height);
+    }
+  else
+    {
+      GDK_DEBUG (VULKAN, "Vulkan %ux%u and GTK %ux%u disagree on size, using GTK's",
+                 (guint) capabilities.currentExtent.width, (guint) capabilities.currentExtent.height,
+                 (guint) size.width, (guint) size.height);
+    }
 
-      capabilities.currentExtent.width = MAX (1, (int) ceil (gdk_surface_get_width (surface) * scale));
-      capabilities.currentExtent.height = MAX (1, (int) ceil (gdk_surface_get_height (surface) * scale));
-
-      GDK_DEBUG (VULKAN, "Effective extent %dx%d",
-                 capabilities.currentExtent.width,
-                 capabilities.currentExtent.height);
+  if (capabilities.minImageExtent.width > size.width ||
+      capabilities.minImageExtent.height > size.height ||
+      capabilities.maxImageExtent.width < size.width ||
+      capabilities.maxImageExtent.height < size.height)
+    {
+      /* Should this be a g_warning()? */
+      GDK_DEBUG (VULKAN, "Clamping size %ux%u to fit into min %ux%u, max %ux%u",
+                 (guint) size.width, (guint) size.height,
+                 (guint) capabilities.minImageExtent.width, (guint) capabilities.minImageExtent.height,
+                 (guint) capabilities.maxImageExtent.width, (guint) capabilities.maxImageExtent.height);
+      size.width = CLAMP (size.width, capabilities.minImageExtent.width, capabilities.maxImageExtent.width);
+      size.height = CLAMP (size.height, capabilities.minImageExtent.height, capabilities.maxImageExtent.height);
     }
 
   res = GDK_VK_CHECK (vkCreateSwapchainKHR, device,
@@ -493,7 +510,7 @@ gdk_vulkan_context_check_swapchain (GdkVulkanContext  *context,
                                                                         capabilities.maxImageCount ? capabilities.maxImageCount : G_MAXUINT32),
                                                 .imageFormat = priv->formats[priv->current_depth].vk_format.format,
                                                 .imageColorSpace = priv->formats[priv->current_depth].vk_format.colorSpace,
-                                                .imageExtent = capabilities.currentExtent,
+                                                .imageExtent = size,
                                                 .imageArrayLayers = 1,
                                                 .imageUsage = VK_IMAGE_USAGE_COLOR_ATTACHMENT_BIT,
                                                 .imageSharingMode = VK_SHARING_MODE_EXCLUSIVE,
@@ -542,8 +559,8 @@ gdk_vulkan_context_check_swapchain (GdkVulkanContext  *context,
         {
           priv->regions[i] = cairo_region_create_rectangle (&(cairo_rectangle_int_t) {
                                                                 0, 0,
-                                                                gdk_surface_get_width (surface),
-                                                                gdk_surface_get_height (surface),
+                                                                size.width,
+                                                                size.height
                                                             });
         }
     }
@@ -577,7 +594,7 @@ physical_device_supports_extension (VkPhysicalDevice  device,
     {
       first = FALSE;
       for (uint32_t i = 0; i < n_device_extensions; i++)
-        g_print ("%s\n", extensions[i].extensionName);
+        GDK_DEBUG (VULKAN, "%s", extensions[i].extensionName);
     }
 
   for (uint32_t i = 0; i < n_device_extensions; i++)
@@ -647,6 +664,9 @@ physical_device_check_features (VkPhysicalDevice device)
       physical_device_supports_extension (device, VK_EXT_SWAPCHAIN_MAINTENANCE_1_EXTENSION_NAME))
     features |= GDK_VULKAN_FEATURE_SWAPCHAIN_MAINTENANCE;
 
+  if (physical_device_supports_extension (device, VK_EXT_SWAPCHAIN_COLOR_SPACE_EXTENSION_NAME))
+    features |= GDK_VULKAN_FEATURE_SWAPCHAIN_COLORSPACE;
+
   return features;
 }
 
@@ -765,15 +785,11 @@ gdk_vulkan_context_end_frame (GdkDrawContext *draw_context,
 {
   GdkVulkanContext *context = GDK_VULKAN_CONTEXT (draw_context);
   GdkVulkanContextPrivate *priv = gdk_vulkan_context_get_instance_private (context);
-  GdkSurface *surface = gdk_draw_context_get_surface (draw_context);
   VkRectLayerKHR *rectangles;
   int n_regions;
 
   if (gdk_vulkan_context_has_feature (context, GDK_VULKAN_FEATURE_INCREMENTAL_PRESENT))
     {
-      double scale;
-
-      scale = gdk_surface_get_scale (surface);
       n_regions = cairo_region_num_rectangles (painted);
       rectangles = g_alloca (sizeof (VkRectLayerKHR) * n_regions);
 
@@ -784,11 +800,10 @@ gdk_vulkan_context_end_frame (GdkDrawContext *draw_context,
           cairo_region_get_rectangle (painted, i, &r);
 
           rectangles[i] = (VkRectLayerKHR) {
-              .layer = 0,
-              .offset.x = (int) floor (r.x * scale),
-              .offset.y = (int) floor (r.y * scale),
-              .extent.width = (int) ceil ((r.x + r.width) * scale) - floor (r.x * scale),
-              .extent.height = (int) ceil ((r.y + r.height) * scale) - floor (r.y * scale),
+              .offset.x = r.x,
+              .offset.y = r.y,
+              .extent.width = r.width,
+              .extent.height = r.height
           };
         }
     }
@@ -887,8 +902,7 @@ gdk_vulkan_context_real_init (GInitable     *initable,
   VkBool32 supported;
   uint32_t i;
 
-  priv->vulkan_ref = gdk_display_init_vulkan (display, error);
-  if (!priv->vulkan_ref)
+  if (!gdk_display_prepare_vulkan (display, error))
     return FALSE;
 
   if (surface == NULL)
@@ -927,6 +941,7 @@ gdk_vulkan_context_real_init (GInitable     *initable,
   else
     {
       uint32_t n_formats;
+
       GDK_VK_CHECK (vkGetPhysicalDeviceSurfaceFormatsKHR, gdk_vulkan_context_get_physical_device (context),
                                                           priv->surface,
                                                           &n_formats, NULL);
@@ -936,13 +951,16 @@ gdk_vulkan_context_real_init (GInitable     *initable,
                                                           &n_formats, formats);
       for (i = 0; i < n_formats; i++)
         {
-          if (formats[i].colorSpace != VK_COLOR_SPACE_SRGB_NONLINEAR_KHR)
+          if (formats[i].colorSpace != VK_COLOR_SPACE_SRGB_NONLINEAR_KHR &&
+              formats[i].colorSpace != VK_COLOR_SPACE_PASS_THROUGH_EXT)
             continue;
 
           switch ((int) formats[i].format)
             {
               case VK_FORMAT_B8G8R8A8_UNORM:
-                if (priv->formats[GDK_MEMORY_U8].vk_format.format == VK_FORMAT_UNDEFINED)
+                if (priv->formats[GDK_MEMORY_U8].vk_format.format == VK_FORMAT_UNDEFINED ||
+                    (formats[i].colorSpace == VK_COLOR_SPACE_PASS_THROUGH_EXT &&
+                     priv->formats[GDK_MEMORY_U16].vk_format.colorSpace != VK_COLOR_SPACE_PASS_THROUGH_EXT))
                   {
                     priv->formats[GDK_MEMORY_U8].vk_format = formats[i];
                     priv->formats[GDK_MEMORY_U8].gdk_format = GDK_MEMORY_B8G8R8A8_PREMULTIPLIED;
@@ -950,7 +968,9 @@ gdk_vulkan_context_real_init (GInitable     *initable,
                 break;
 
               case VK_FORMAT_R8G8B8A8_UNORM:
-                if (priv->formats[GDK_MEMORY_U8].vk_format.format == VK_FORMAT_UNDEFINED)
+                if (priv->formats[GDK_MEMORY_U8].vk_format.format == VK_FORMAT_UNDEFINED ||
+                    (formats[i].colorSpace == VK_COLOR_SPACE_PASS_THROUGH_EXT &&
+                     priv->formats[GDK_MEMORY_U16].vk_format.colorSpace != VK_COLOR_SPACE_PASS_THROUGH_EXT))
                   {
                     priv->formats[GDK_MEMORY_U8].vk_format = formats[i];
                     priv->formats[GDK_MEMORY_U8].gdk_format = GDK_MEMORY_R8G8B8A8_PREMULTIPLIED;
@@ -958,7 +978,9 @@ gdk_vulkan_context_real_init (GInitable     *initable,
                 break;
 
               case VK_FORMAT_B8G8R8A8_SRGB:
-                if (priv->formats[GDK_MEMORY_U8_SRGB].vk_format.format == VK_FORMAT_UNDEFINED)
+                if (priv->formats[GDK_MEMORY_U8_SRGB].vk_format.format == VK_FORMAT_UNDEFINED ||
+                    (formats[i].colorSpace == VK_COLOR_SPACE_PASS_THROUGH_EXT &&
+                     priv->formats[GDK_MEMORY_U16].vk_format.colorSpace != VK_COLOR_SPACE_PASS_THROUGH_EXT))
                   {
                     priv->formats[GDK_MEMORY_U8_SRGB].vk_format = formats[i];
                     priv->formats[GDK_MEMORY_U8_SRGB].gdk_format = GDK_MEMORY_B8G8R8A8_PREMULTIPLIED;
@@ -966,7 +988,9 @@ gdk_vulkan_context_real_init (GInitable     *initable,
                 break;
 
               case VK_FORMAT_R8G8B8A8_SRGB:
-                if (priv->formats[GDK_MEMORY_U8_SRGB].vk_format.format == VK_FORMAT_UNDEFINED)
+                if (priv->formats[GDK_MEMORY_U8_SRGB].vk_format.format == VK_FORMAT_UNDEFINED ||
+                    (formats[i].colorSpace == VK_COLOR_SPACE_PASS_THROUGH_EXT &&
+                     priv->formats[GDK_MEMORY_U16].vk_format.colorSpace != VK_COLOR_SPACE_PASS_THROUGH_EXT))
                   {
                     priv->formats[GDK_MEMORY_U8_SRGB].vk_format = formats[i];
                     priv->formats[GDK_MEMORY_U8_SRGB].gdk_format = GDK_MEMORY_R8G8B8A8_PREMULTIPLIED;
@@ -974,18 +998,27 @@ gdk_vulkan_context_real_init (GInitable     *initable,
                 break;
 
               case VK_FORMAT_R16G16B16A16_UNORM:
-                priv->formats[GDK_MEMORY_U16].vk_format = formats[i];
-                priv->formats[GDK_MEMORY_U16].gdk_format = GDK_MEMORY_R16G16B16A16_PREMULTIPLIED;
+                if (priv->formats[GDK_MEMORY_U16].vk_format.colorSpace != VK_COLOR_SPACE_PASS_THROUGH_EXT)
+                  {
+                    priv->formats[GDK_MEMORY_U16].vk_format = formats[i];
+                    priv->formats[GDK_MEMORY_U16].gdk_format = GDK_MEMORY_R16G16B16A16_PREMULTIPLIED;
+                  }
                 break;
 
               case VK_FORMAT_R16G16B16A16_SFLOAT:
-                priv->formats[GDK_MEMORY_FLOAT16].vk_format = formats[i];
-                priv->formats[GDK_MEMORY_FLOAT16].gdk_format = GDK_MEMORY_R16G16B16A16_FLOAT_PREMULTIPLIED;
+                if (priv->formats[GDK_MEMORY_FLOAT16].vk_format.colorSpace != VK_COLOR_SPACE_PASS_THROUGH_EXT)
+                  {
+                    priv->formats[GDK_MEMORY_FLOAT16].vk_format = formats[i];
+                    priv->formats[GDK_MEMORY_FLOAT16].gdk_format = GDK_MEMORY_R16G16B16A16_FLOAT_PREMULTIPLIED;
+                  }
                 break;
 
               case VK_FORMAT_R32G32B32A32_SFLOAT:
-                priv->formats[GDK_MEMORY_FLOAT32].vk_format = formats[i];
-                priv->formats[GDK_MEMORY_FLOAT32].gdk_format = GDK_MEMORY_R32G32B32A32_FLOAT_PREMULTIPLIED;
+                if (priv->formats[GDK_MEMORY_FLOAT32].vk_format.colorSpace != VK_COLOR_SPACE_PASS_THROUGH_EXT)
+                  {
+                    priv->formats[GDK_MEMORY_FLOAT32].vk_format = formats[i];
+                    priv->formats[GDK_MEMORY_FLOAT32].gdk_format = GDK_MEMORY_R32G32B32A32_FLOAT_PREMULTIPLIED;
+                  }
                 break;
 
               default:
@@ -1455,9 +1488,6 @@ gdk_display_create_vulkan_device (GdkDisplay  *display,
 {
   G_GNUC_UNUSED gint64 start_time = GDK_PROFILER_CURRENT_TIME;
   uint32_t i, j, k;
-  const char *override;
-  gboolean list_devices;
-  int first, last;
   GdkVulkanFeatures skip_features;
 
   uint32_t n_devices = 0;
@@ -1475,9 +1505,6 @@ gdk_display_create_vulkan_device (GdkDisplay  *display,
   devices = g_newa (VkPhysicalDevice, n_devices);
   GDK_VK_CHECK(vkEnumeratePhysicalDevices, display->vk_instance, &n_devices, devices);
 
-  first = 0;
-  last = n_devices;
-
   skip_features = gdk_parse_debug_var ("GDK_VULKAN_DISABLE",
       "GDK_VULKAN_DISABLE can be set to a list of Vulkan features to disable.\n",
       gdk_vulkan_feature_keys,
@@ -1485,35 +1512,7 @@ gdk_display_create_vulkan_device (GdkDisplay  *display,
   if (skip_features & GDK_VULKAN_FEATURE_YCBCR)
     skip_features |= GDK_VULKAN_FEATURE_DMABUF;
 
-  override = g_getenv ("GDK_VULKAN_DEVICE");
-  list_devices = FALSE;
-  if (override)
-    {
-      if (g_strcmp0 (override, "list") == 0)
-        list_devices = TRUE;
-      else
-        {
-          gint64 device_idx;
-          GError *error2 = NULL;
-
-          if (!g_ascii_string_to_signed (override, 10, 0, G_MAXINT, &device_idx, &error2))
-            {
-              g_warning ("Failed to parse %s: %s", "GDK_VULKAN_DEVICE", error2->message);
-              g_error_free (error2);
-              device_idx = -1;
-            }
-
-          if (device_idx < 0 || device_idx >= n_devices)
-            g_warning ("%s value out of range, ignoring", "GDK_VULKAN_DEVICE");
-          else
-            {
-              first = device_idx;
-              last = first + 1;
-            }
-        }
-    }
-
-  if (list_devices || GDK_DISPLAY_DEBUG_CHECK (display, VULKAN))
+  if (GDK_DISPLAY_DEBUG_CHECK (display, VULKAN))
     {
       for (i = 0; i < n_devices; i++)
         {
@@ -1564,7 +1563,7 @@ gdk_display_create_vulkan_device (GdkDisplay  *display,
         }
     }
 
-  for (i = first; i < last; i++)
+  for (i = 0; i < n_devices; i++)
     {
       GdkVulkanFeatures features, device_features;
       uint32_t n_queue_props;
@@ -1580,7 +1579,7 @@ gdk_display_create_vulkan_device (GdkDisplay  *display,
         {
           if (queue_props[j].queueFlags & VK_QUEUE_GRAPHICS_BIT)
             {
-              VkPhysicalDeviceSwapchainMaintenance1FeaturesEXT swapchain_maintenance1_features = { 
+              VkPhysicalDeviceSwapchainMaintenance1FeaturesEXT swapchain_maintenance1_features = {
                 .sType = VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SWAPCHAIN_MAINTENANCE_1_FEATURES_EXT,
                 .swapchainMaintenance1 = VK_TRUE,
               };
@@ -1619,6 +1618,8 @@ gdk_display_create_vulkan_device (GdkDisplay  *display,
                   swapchain_maintenance1_features.pNext = create_device_pNext;
                   create_device_pNext = &swapchain_maintenance1_features;
                 }
+              if (features & GDK_VULKAN_FEATURE_SWAPCHAIN_COLORSPACE)
+                g_ptr_array_add (device_extensions, (gpointer) VK_EXT_SWAPCHAIN_COLOR_SPACE_EXTENSION_NAME);
 
 #define ENABLE_IF(flag) ((features & (flag)) ? VK_TRUE : VK_FALSE)
               GDK_DISPLAY_DEBUG (display, VULKAN, "Using Vulkan device %u, queue %u", i, j);
@@ -1678,7 +1679,7 @@ gdk_display_create_vulkan_device (GdkDisplay  *display,
   return FALSE;
 }
 
-static VkBool32 VKAPI_CALL
+static VKAPI_ATTR VkBool32 VKAPI_CALL
 gdk_vulkan_debug_report (VkDebugReportFlagsEXT      flags,
                          VkDebugReportObjectTypeEXT objectType,
                          uint64_t                   object,
@@ -1702,7 +1703,7 @@ gdk_vulkan_debug_report (VkDebugReportFlagsEXT      flags,
   return VK_FALSE;
 }
 
-static gboolean
+gboolean
 gdk_display_create_vulkan_instance (GdkDisplay  *display,
                                     GError     **error)
 {
@@ -1712,6 +1713,8 @@ gdk_display_create_vulkan_instance (GdkDisplay  *display,
   gboolean have_debug_report = FALSE;
   VkResult res;
 
+  g_assert (display->vk_instance == NULL);
+
   if (!gdk_has_feature (GDK_FEATURE_VULKAN))
     {
       g_set_error_literal (error, GDK_VULKAN_ERROR, GDK_VULKAN_ERROR_NOT_AVAILABLE,
@@ -1838,66 +1841,16 @@ gdk_display_create_vulkan_instance (GdkDisplay  *display,
   return TRUE;
 }
 
-/*
- * gdk_display_init_vulkan:
- * @display: a display
- * @error: A potential error message
- *
- * Initializes Vulkan and returns an error on failure.
- *
- * If Vulkan is already initialized, this function returns
- * %TRUE and increases the refcount of the existing instance.
- *
- * You need to gdk_display_unref_vulkan() to close it again.
- *
- * Returns: %TRUE if Vulkan is initialized.
- **/
-gboolean
-gdk_display_init_vulkan (GdkDisplay *display,
-                         GError     **error)
-{
-  if (display->vulkan_refcount == 0)
-    {
-      if (!gdk_display_create_vulkan_instance (display, error))
-        return FALSE;
-    }
-
-  display->vulkan_refcount++;
-
-  return TRUE;
-}
-
-/*
- * gdk_display_ref_vulkan:
- * @display: a GdkDisplay
- *
- * Increases the refcount of an existing Vulkan instance.
- *
- * This function must not be called if Vulkan may not be initialized
- * yet, call gdk_display_init_vulkan() in that case.
- **/
 void
-gdk_display_ref_vulkan (GdkDisplay *display)
-{
-  g_assert (display->vulkan_refcount > 0);
-
-  display->vulkan_refcount++;
-}
-
-void
-gdk_display_unref_vulkan (GdkDisplay *display)
+gdk_display_destroy_vulkan_instance (GdkDisplay *display)
 {
   GHashTableIter iter;
   gpointer key, value;
 
-  g_return_if_fail (GDK_IS_DISPLAY (display));
-  g_return_if_fail (display->vulkan_refcount > 0);
+  g_assert (GDK_IS_DISPLAY (display));
+  g_assert (display->vk_instance != NULL);
 
-  display->vulkan_refcount--;
-  if (display->vulkan_refcount > 0)
-    return;
-
-  GDK_DEBUG (VULKAN, "Closing Vulkan instance");
+  GDK_DEBUG (VULKAN, "Destroy Vulkan instance");
   display->vulkan_features = 0;
   g_clear_pointer (&display->vk_dmabuf_formats, gdk_dmabuf_formats_unref);
   g_hash_table_iter_init (&iter, display->vk_shader_modules);
@@ -1975,11 +1928,9 @@ gdk_vulkan_init_dmabuf (GdkDisplay *display)
     return;
 
   if (!gdk_has_feature (GDK_FEATURE_DMABUF) ||
-      !gdk_display_init_vulkan (display, NULL) ||
-      ((display->vulkan_features & GDK_VULKAN_FEATURE_DMABUF) == 0))
-    {
-      return;
-    }
+      !gdk_display_prepare_vulkan (display, NULL) ||
+      (display->vulkan_features & GDK_VULKAN_FEATURE_DMABUF) == 0)
+    return;
 
   vulkan_builder = gdk_dmabuf_formats_builder_new ();
 
@@ -1988,26 +1939,27 @@ gdk_vulkan_init_dmabuf (GdkDisplay *display)
       if (vk_format == VK_FORMAT_UNDEFINED)
         continue;
 
-      modifier_props.drmFormatModifierCount = sizeof (modifier_list);
+      modifier_props.drmFormatModifierCount = G_N_ELEMENTS (modifier_list);
       vkGetPhysicalDeviceFormatProperties2 (display->vk_physical_device,
                                             vk_format,
                                             &props);
-      g_warn_if_fail (modifier_props.drmFormatModifierCount < sizeof (modifier_list));
+      g_warn_if_fail (modifier_props.drmFormatModifierCount < G_N_ELEMENTS (modifier_list));
       for (j = 0; j < modifier_props.drmFormatModifierCount; j++)
         {
+          gboolean advertise = modifier_list[j].drmFormatModifier != DRM_FORMAT_MOD_LINEAR;
+
           GDK_DISPLAY_DEBUG (display, DMABUF,
-                             "Vulkan supports dmabuf format %.4s::%016llx with %u planes and features 0x%x",
+                             "Vulkan %s dmabuf format %.4s::%016"G_GINT64_MODIFIER"x with %u planes and features 0x%x",
+                             advertise ? "advertises" : "supports",
                              (char *) &fourcc,
-                             (long long unsigned) modifier_list[j].drmFormatModifier,
+                             modifier_list[j].drmFormatModifier,
                              modifier_list[j].drmFormatModifierPlaneCount,
                              modifier_list[j].drmFormatModifierTilingFeatures);
 
-          if (modifier_list[j].drmFormatModifier == DRM_FORMAT_MOD_LINEAR)
-            continue;
-
-          gdk_dmabuf_formats_builder_add_format (vulkan_builder,
-                                                 fourcc,
-                                                 modifier_list[j].drmFormatModifier);
+          if (advertise)
+            gdk_dmabuf_formats_builder_add_format (vulkan_builder,
+                                                   fourcc,
+                                                   modifier_list[j].drmFormatModifier);
         }
     }
 
